/* Peter Senna Tschudin - peter.senna@gmail.com */

#include "rdtscbench.h"
#include <memory.h>

#define X_SIZE	60
#define Y_SIZE	30


int matrix[X_SIZE][Y_SIZE];

int main (int argc, char *argv[]){
	struct timespec begin, end;
	int accrepeats, repeats, detailed = 0;

	if ( argc < 3 ){
		printf ("Error. You need to specify two numbers as arguments.\n\n %s <repeats> <acc repeats> <optional detailed run parameter. Try 1>\n\n" , argv[0]);
		exit (1);
	}
	repeats = atoi(argv[1]);
	accrepeats = atoi(argv[2]);

	if ( argc > 3 )
		detailed = 1;

	measure ( "matrix1()       ", repeats, detailed, &MATRIX1 );
	sleep(5);

	measure ( "matrix2()       ", repeats, detailed, &MATRIX2 );
	sleep(5);

	measure ( "matrix3()       ", repeats, detailed, &MATRIX3 );
	sleep(5);

	measure ( "matrix4()       ", repeats, detailed, &MATRIX4 );
	sleep(5);

	measure ( "matrix5()       ", repeats, detailed, &MATRIX5 );
	sleep(5);

	measure ( "matrix6()       ", repeats, detailed, &MATRIX6 );
	sleep(5);

	measure ( "matrix7()       ", repeats, detailed, &MATRIX7 );
	sleep(5);

	measure ( "matrix8()       ", repeats, detailed, &MATRIX8 );
	sleep(5);

	accuracy (accrepeats, detailed );

}

/*
 * MATRIX 1
 */

void inline MATRIX1 (struct timespec *ts, struct timespec *te){
	#define X_SIZE	60
	#define Y_SIZE	30

	int x, y;

	/* ### starts measuring ### */
	getticks ( ts );

	for (x = 0; x < X_SIZE; ++x){
		for (y = 0; y < Y_SIZE; ++y){
			matrix[x][y] = -1;
		}
	}

	/* ### ends measuring ### */
	getticks ( te );

}

/*
 * MATRIX 2
 */
void inline MATRIX2 (struct timespec *ts, struct timespec *te){
	#define X_SIZE	60
	#define Y_SIZE	30

	register int x,y;

	/* ### starts measuring ### */
	getticks ( ts );

	for (x = 0; x < X_SIZE; ++x){
		for (y = 0; y < Y_SIZE; ++y){
			matrix[x][y] = -1;
		}
	}

	/* ### ends measuring ### */
	getticks ( te );

}

/*
 * MATRIX 3
 */
void inline MATRIX3 (struct timespec *ts, struct timespec *te){
	#define X_SIZE	60
	#define Y_SIZE	30

	register int x,y;

	/* ### starts measuring ### */
	getticks ( ts );

	for (y = 0; y < Y_SIZE; ++y){
		for (x = 0; x < X_SIZE; ++x){
			matrix[x][y] = -1;
		}
	}

	/* ### ends measuring ### */
	getticks ( te );

}

/*
 * MATRIX 4
 */
void inline MATRIX4 (struct timespec *ts, struct timespec *te){
	#define X_SIZE	60
	#define Y_SIZE	32

	register int x,y;

	/* ### starts measuring ### */
	getticks ( ts );

	for (y = 0; y < Y_SIZE; ++y){
		for (x = 0; x < X_SIZE; ++x){
			matrix[x][y] = -1;
		}
	}

	/* ### ends measuring ### */
	getticks ( te );

}

/*
 * MATRIX 5
 */
void inline MATRIX5 (struct timespec *ts, struct timespec *te){
	#define X_SIZE	60
	#define Y_SIZE	30

	register int index;
	register int *matrix_ptr;

	/* ### starts measuring ### */
	getticks ( ts );

	matrix_ptr = &matrix[0][0];

	for (index = 0; index < X_SIZE * Y_SIZE; ++index){
		*matrix_ptr = -1;
		matrix_ptr++;
	}

	/* ### ends measuring ### */
	getticks ( te );

}

/*
 * MATRIX 6
 */
void inline MATRIX6 (struct timespec *ts, struct timespec *te){
	#define X_SIZE	60
	#define Y_SIZE	30

	register int *matrix_ptr;

	/* ### starts measuring ### */
	getticks ( ts );

	for (matrix_ptr = &matrix[0][0];
		matrix_ptr <= &matrix[X_SIZE - 1][Y_SIZE - 1];
		++matrix_ptr){
		*matrix_ptr = -1;
	}

	/* ### ends measuring ### */
	getticks ( te );

}

/*
 * MATRIX 7 
 */
void initmatrix(){
	memset (matrix, -1, sizeof(matrix));
}

void inline MATRIX7 (struct timespec *ts, struct timespec *te){
	#define X_SIZE	60
	#define Y_SIZE	30

	register int *matrix_ptr;

	/* ### starts measuring ### */
	getticks ( ts );

	initmatrix();

	/* ### ends measuring ### */
	getticks ( te );

}

/*
 * MATRIX 8
 */
#define init_matrix() \
	memset (matrix, -1, sizeof(matrix));

void inline MATRIX8 (struct timespec *ts, struct timespec *te){
	#define X_SIZE	60
	#define Y_SIZE	30

	register int *matrix_ptr;

	/* ### starts measuring ### */
	getticks ( ts );

	init_matrix();

	/* ### ends measuring ### */
	getticks ( te );

}

/* 
 * rdtsc internal functions 
 */

/* assembly code to read the TSC */
static __inline__ void getticks(struct timespec *ts) {
	unsigned int a, d;

	asm("cpuid");
	asm volatile("rdtsc" : "=a" (a), "=d" (d));

	ts->tv_nsec = (((uint64_t)a) | (((uint64_t)d) << 32));
}

/* Accuracy detection */
long double accuracy ( int repeats, int detailed ){
	struct timespec begin, end, diff;
	long double accuracy = 0, acc = 1;
	int i;

	printf ("\n----------------\n\naccuracy       :");

	for ( i = repeats; i > 0; i-- ){

		getticks ( &begin );
		sleep ( 1 );
		getticks ( &end );

		diff.tv_nsec = (long)(double) (end.tv_nsec - begin.tv_nsec);
		acc *= diff.tv_nsec;
		if ( detailed == 1 ){
			accuracy = diff.tv_nsec;
			accuracy *= 100;
			accuracy /= CPU_HZ;
			if ( accuracy > 100 ){
				accuracy -= 100;
				accuracy = 100 - accuracy;
			}
			printf ( "%2.6Lf, ", accuracy );
		}
	}
	long double power = 1;
	power /= repeats;

	accuracy = powl ( acc, power );

	accuracy *= 100;
	accuracy /= CPU_HZ;

	if ( accuracy > 100 ){
		accuracy -= 100;
		accuracy = 100 - accuracy;
	}

	printf ("  %25.6Lf%%       Geom. mean of %4.0d values.\n\n", accuracy, repeats );

	return accuracy;
}

/* General tool to measure run time of *function */
void inline measure (char msg[32], int repeats, int detailed, void (*function)( struct timespec *ts, struct timespec *te ) ){

	struct timespec begin, end, diff;
	long double acc = 1;
	int i = 0;

	printf ("%s: ", msg);

	for ( i = repeats; i > 0; i-- ){

		//getticks ( &begin );
		function( &begin, &end );
		//getticks ( &end );

		diff.tv_nsec = end.tv_nsec - begin.tv_nsec;

		acc *= (long)(double)diff.tv_nsec;
		if ( detailed == 1 )
			printf ( " %llu, ", diff.tv_nsec );
	}
	long double power = 1;
	power /= repeats;

	long double valuemean = powl ( acc, power );
	printf ("%25.6Lf clocks. Geom. mean of %4.0d values.\n", valuemean, repeats );
}
